# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.common import constants as n_const
from neutron.plugins.common import constants as p_const


CLOUD_PLATFORM_NAME = 'OpenStack'

AGENT_BINARY_NAME = 'infoblox-ipam-agent'
AGENT_TYPE_INFOBLOX_IPAM = 'Infoblox IPAM agent'
AGENT_TOPIC = 'infoblox_ipam_agent'
AGENT_NOTIFICATION_POOL = 'infoblox-ipam-notification'

NOTIFICATION_EXCHANGE_NEUTRON = 'neutron'
NOTIFICATION_EXCHANGE_NOVA = 'nova'

GRID_STATUS_ON = 'ON'
GRID_STATUS_OFF = 'OFF'

MEMBER_RESERVATION_RETRY = 5
MEMBER_STATUS_ON = 'ON'
MEMBER_STATUS_OFF = 'OFF'

MEMBER_NODE_STATUS_FAILED = 'FAILED'
MEMBER_NODE_STATUS_INACTIVE = 'INACTIVE'
MEMBER_NODE_STATUS_WARNING = 'WARNING'
MEMBER_NODE_STATUS_WORKING = 'WORKING'

MEMBER_TYPE_GRID_MASTER = 'GM'
MEMBER_TYPE_CP_MEMBER = 'CPM'
MEMBER_TYPE_REGULAR_MEMBER = 'REGULAR'

# Cloud Network Automation for GM
MEMBER_LICENSE_TYPE_CLOUD = 'CLOUD'
# Cloud platform for CPM
MEMBER_LICENSE_TYPE_CLOUD_API = 'CLOUD_API'

SERVICE_TYPE_DHCP = 'DHCP'
SERVICE_TYPE_DNS = 'DNS'

EA_GRID_CONFIG_GRID_SYNC_SUPPORT = 'Grid Sync Support'
EA_GRID_CONFIG_GRID_SYNC_MINIMUM_WAIT_TIME = 'Grid Sync Minimum Wait Time'
EA_GRID_CONFIG_GRID_SYNC_MAXIMUM_WAIT_TIME = 'Grid Sync Maximum Wait Time'
EA_GRID_CONFIG_DEFAULT_NETWORK_VIEW_SCOPE = 'Default Network View Scope'
EA_GRID_CONFIG_DEFAULT_NETWORK_VIEW = 'Default Network View'
EA_GRID_CONFIG_DEFAULT_HOST_NAME_PATTERN = 'Default Host Name Pattern'
EA_GRID_CONFIG_DEFAULT_DOMAIN_NAME_PATTERN = 'Default Domain Name Pattern'
EA_GRID_CONFIG_NS_GROUP = 'NS Group'
EA_GRID_CONFIG_DNS_VIEW = 'DNS View'
EA_GRID_CONFIG_NETWORK_TEMPLATE = 'Network Template'
EA_GRID_CONFIG_ADMIN_NETWORK_DELETION = 'Admin Network Deletion'
EA_GRID_CONFIG_IP_ALLOCATION_STRATEGY = 'IP Allocation Strategy'
EA_GRID_CONFIG_DNS_RECORD_BINDING_TYPES = 'DNS Record Binding Types'
EA_GRID_CONFIG_DNS_RECORD_UNBINDING_TYPES = 'DNS Record Unbinding Types'
EA_GRID_CONFIG_DNS_RECORD_REMOVABLE_TYPES = 'DNS Record Removable Types'
EA_GRID_CONFIG_DHCP_SUPPORT = 'DHCP Support'
EA_GRID_CONFIG_DNS_SUPPORT = 'DNS Support'
EA_GRID_CONFIG_RELAY_SUPPORT = 'Relay Support'
EA_GRID_CONFIG_USE_GM_FOR_DHCP = 'Use Grid Master for DHCP'
EA_GRID_CONFIG_TENANT_NAME_PERSISTENCE = 'Tenant Name Persistence'
EA_GRID_CONFIG_REPORT_GRID_SYNC_TIME = 'Report Grid Sync Time'

EA_LAST_GRID_SYNC_TIME = 'Last Grid Sync Time'

EA_MAPPING_ADDRESS_SCOPE_ID = 'Address Scope ID Mapping'
EA_MAPPING_ADDRESS_SCOPE_NAME = 'Address Scope Name Mapping'
EA_MAPPING_TENANT_ID = 'Tenant ID Mapping'
EA_MAPPING_TENANT_NAME = 'Tenant Name Mapping'
EA_MAPPING_NETWORK_ID = 'Network ID Mapping'
EA_MAPPING_NETWORK_NAME = 'Network Name Mapping'
EA_MAPPING_SUBNET_ID = 'Subnet ID Mapping'
EA_MAPPING_SUBNET_CIDR = 'Subnet CIDR Mapping'

EA_CLOUD_ADAPTER_ID = 'Cloud Adapter ID'
EA_IS_CLOUD_MEMBER = 'Is Cloud Member'

EA_SUBNET_ID = 'Subnet ID'
EA_SUBNET_NAME = 'Subnet Name'
EA_NETWORK_ID = 'Network ID'
EA_NETWORK_NAME = 'Network Name'
EA_NETWORK_ENCAP = 'Network Encap'
EA_SEGMENTATION_ID = 'Segmentation ID'
EA_PHYSICAL_NETWORK_NAME = 'Physical Network Name'
EA_PORT_ID = 'Port ID'
EA_PORT_DEVICE_OWNER = 'Port Attached Device - Device Owner'
EA_PORT_DEVICE_ID = 'Port Attached Device - Device ID'
EA_VM_ID = 'VM ID'
EA_VM_NAME = 'VM Name'
EA_IP_TYPE = 'IP Type'
EA_TENANT_ID = 'Tenant ID'
EA_TENANT_NAME = 'Tenant Name'
EA_ACCOUNT = 'Account'
EA_CLOUD_API_OWNED = 'Cloud API Owned'
EA_CMP_TYPE = 'CMP Type'
EA_IS_EXTERNAL = 'Is External'
EA_IS_SHARED = 'Is Shared'

REQUIRED_EA_LIST = [EA_CMP_TYPE, EA_TENANT_ID, EA_CLOUD_API_OWNED]

NETWORK_EA_LIST = [EA_ACCOUNT, EA_TENANT_NAME,
                   EA_IS_EXTERNAL, EA_IS_SHARED, EA_SUBNET_ID, EA_SUBNET_NAME,
                   EA_NETWORK_ID, EA_NETWORK_NAME, EA_NETWORK_ENCAP,
                   EA_SEGMENTATION_ID, EA_PHYSICAL_NETWORK_NAME]

RANGE_EA_LIST = [EA_ACCOUNT, EA_TENANT_NAME]

ZONE_EA_LIST = [EA_ACCOUNT, EA_TENANT_NAME]

IP_TYPE_FIXED = 'Fixed'
IP_TYPE_FLOATING = 'Floating'
IP_TYPE_PUBLIC = 'Public'
IP_TYPE_PRIVATE = 'Private'
IP_TYPE_ELASTIC = 'Elastic'

IP_ALLOCATION_STRATEGY_HOST_RECORD = 'Host Record'
IP_ALLOCATION_STRATEGY_FIXED_ADDRESS = 'Fixed Address'

DNS_RECORD_TYPE_A = 'record:a'
DNS_RECORD_TYPE_AAAA = 'record:aaaa'
DNS_RECORD_TYPE_PTR = 'record:ptr'
DNS_RECORD_TYPE_TXT = 'record:txt'

NETWORK_VIEW_SCOPE_SINGLE = 'Single'
NETWORK_VIEW_SCOPE_ADDRESS_SCOPE = 'Address Scope'
NETWORK_VIEW_SCOPE_TENANT = 'Tenant'
NETWORK_VIEW_SCOPE_NETWORK = 'Network'
NETWORK_VIEW_SCOPE_SUBNET = 'Subnet'

MAPPING_CONDITION_KEY_NAME = 'neutron_object_name'
MAPPING_CONDITION_VALUE_NAME = 'neutron_object_value'

MAPPING_RELATION_GM_OWNED = 'GM-OWNED'
MAPPING_RELATION_DELEGATED = 'DELEGATED'

DEFAULT_NETWORK_VIEW = 'default'
DEFAULT_DNS_VIEW = 'default'

IS_DEFAULT = 'is_default'

NEUTRON_DEVICE_OWNER_TO_PATTERN_MAP = {
    n_const.DEVICE_OWNER_DHCP: 'dhcp-port-{ip_address}',
    n_const.DEVICE_OWNER_ROUTER_INTF: 'router-iface-{ip_address}',
    n_const.DEVICE_OWNER_ROUTER_GW: 'router-gw-{ip_address}',
    n_const.DEVICE_OWNER_FLOATINGIP: 'floating-ip-{ip_address}',
    'neutron:' + p_const.LOADBALANCER: 'lb-vip-{ip_address}'
}

NEUTRON_INTERNAL_SERVICE_DEVICE_OWNERS = [
    n_const.DEVICE_OWNER_DHCP,
    n_const.DEVICE_OWNER_ROUTER_INTF,
    n_const.DEVICE_OWNER_ROUTER_GW,
    'neutron:' + p_const.LOADBALANCER
]

NEUTRON_DEVICE_OWNER_COMPUTE_NOVA = 'compute:nova'
# nova has a bug that it returns compute:None when an instance is created from
# CLI command: nova-boot.
NEUTRON_DEVICE_OWNER_COMPUTE_NONE = 'compute:None'
NEUTRON_DEVICE_OWNER_COMPUTE_LIST = [NEUTRON_DEVICE_OWNER_COMPUTE_NOVA,
                                     NEUTRON_DEVICE_OWNER_COMPUTE_NONE]

NEUTRON_FLOATING_IP_DEVICE_OWNERS = [
    n_const.DEVICE_OWNER_FLOATINGIP,
    NEUTRON_DEVICE_OWNER_COMPUTE_NOVA,
    NEUTRON_DEVICE_OWNER_COMPUTE_NONE
]

NONE_ID = 'NONE'
EA_RESET_VALUE = 'N/A'

REQUIRED_EA_DEFS = [
    # Cloud Property EAs
    {'name': EA_ACCOUNT, 'type': 'STRING', 'flags': 'C',
     'comment': 'User ID in OpenStack'},

    {'name': EA_CLOUD_API_OWNED, 'type': 'ENUM', 'flags': 'C',
     'list_values': [{'value': 'True'}, {'value': 'False'}],
     'comment': 'Is Cloud API owned'},

    {'name': EA_CMP_TYPE, 'type': 'STRING', 'flags': 'C',
     'comment': 'CMP Types (OpenStack)'},

    {'name': EA_IS_EXTERNAL, 'type': 'ENUM', 'flags': 'C',
     'list_values': [{'value': 'True'}, {'value': 'False'}],
     'comment': 'For networks and network containers only'},

    {'name': EA_IS_SHARED, 'type': 'ENUM', 'flags': 'C',
     'list_values': [{'value': 'True'}, {'value': 'False'}],
     'comment': 'For networks and network containers only'},

    {'name': EA_IP_TYPE, 'type': 'ENUM', 'flags': 'C',
     'list_values': [{'value': IP_TYPE_ELASTIC},
                     {'value': IP_TYPE_FIXED},
                     {'value': IP_TYPE_FLOATING},
                     {'value': IP_TYPE_PRIVATE},
                     {'value': IP_TYPE_PUBLIC}],
     'comment': 'Type of IP address'},

    {'name': EA_NETWORK_ENCAP, 'type': 'STRING', 'flags': 'C',
     'comment': 'Type of IP address'},

    {'name': EA_NETWORK_ID, 'type': 'STRING', 'flags': 'C',
     'comment': 'Network ID in OpenStack'},

    {'name': EA_NETWORK_NAME, 'type': 'STRING', 'flags': 'C',
     'comment': 'Network Name'},

    {'name': EA_PHYSICAL_NETWORK_NAME, 'type': 'STRING', 'flags': 'C',
     'comment': ''},

    {'name': EA_PORT_DEVICE_ID, 'type': 'STRING', 'flags': 'C',
     'comment': ''},

    {'name': EA_PORT_DEVICE_OWNER, 'type': 'STRING', 'flags': 'C',
     'comment': ''},

    {'name': EA_PORT_ID, 'type': 'STRING', 'flags': 'C',
     'comment': ''},

    {'name': EA_SEGMENTATION_ID, 'type': 'STRING', 'flags': 'C',
     'comment': ''},

    {'name': EA_SUBNET_ID, 'type': 'STRING', 'flags': 'C',
     'comment': 'Subnet ID in OpenStack'},

    {'name': EA_SUBNET_NAME, 'type': 'STRING', 'flags': 'C',
     'comment': 'Subnet Name in OpenStack'},

    {'name': EA_TENANT_ID, 'type': 'STRING', 'flags': 'C',
     'comment': 'Tenant ID in OpenStack'},

    {'name': EA_TENANT_NAME, 'type': 'STRING', 'flags': 'C',
     'comment': 'Tenant Name in OpenStack'},

    {'name': EA_VM_ID, 'type': 'STRING', 'flags': 'C',
     'comment': 'Instance ID in OpenStack'},

    {'name': EA_VM_NAME, 'type': 'STRING', 'flags': 'C',
     'comment': 'Instance Name in OpenStack'},

    # Grid Configuration EAs
    {'name': EA_GRID_CONFIG_GRID_SYNC_SUPPORT,
     'type': 'ENUM', 'flags': '',
     'allowed_object_types': ['Member'],
     'list_values': [{'value': 'True'},
                     {'value': 'False'}],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_GRID_SYNC_MINIMUM_WAIT_TIME,
     'type': 'INTEGER', 'flags': '',
     'allowed_object_types': ['Member'],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_GRID_SYNC_MAXIMUM_WAIT_TIME,
     'type': 'INTEGER', 'flags': '',
     'allowed_object_types': ['Member'],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_DEFAULT_NETWORK_VIEW_SCOPE,
     'type': 'ENUM', 'flags': '',
     'allowed_object_types': ['Member'],
     'list_values': [{'value': NETWORK_VIEW_SCOPE_SINGLE},
                     {'value': NETWORK_VIEW_SCOPE_ADDRESS_SCOPE},
                     {'value': NETWORK_VIEW_SCOPE_TENANT},
                     {'value': NETWORK_VIEW_SCOPE_NETWORK},
                     {'value': NETWORK_VIEW_SCOPE_SUBNET}],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_DEFAULT_NETWORK_VIEW,
     'type': 'STRING', 'flags': '',
     'allowed_object_types': ['Member'],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_DEFAULT_HOST_NAME_PATTERN,
     'type': 'STRING', 'flags': '',
     'allowed_object_types': ['Member'],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_DEFAULT_DOMAIN_NAME_PATTERN,
     'type': 'STRING', 'flags': '',
     'allowed_object_types': ['Member'],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_NS_GROUP,
     'type': 'STRING', 'flags': '',
     'allowed_object_types': ['Member'],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_DNS_VIEW,
     'type': 'STRING', 'flags': '',
     'allowed_object_types': ['Member'],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_NETWORK_TEMPLATE,
     'type': 'STRING', 'flags': '',
     'allowed_object_types': ['Member'],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_ADMIN_NETWORK_DELETION,
     'type': 'ENUM', 'flags': '',
     'allowed_object_types': ['Member'],
     'list_values': [{'value': 'True'},
                     {'value': 'False'}],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_IP_ALLOCATION_STRATEGY,
     'type': 'ENUM', 'flags': '',
     'allowed_object_types': ['Member'],
     'list_values': [{'value': IP_ALLOCATION_STRATEGY_HOST_RECORD},
                     {'value': IP_ALLOCATION_STRATEGY_FIXED_ADDRESS}],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_DNS_RECORD_BINDING_TYPES,
     'type': 'STRING', 'flags': 'V',
     'allowed_object_types': ['Member'],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_DNS_RECORD_UNBINDING_TYPES,
     'type': 'STRING', 'flags': 'V',
     'allowed_object_types': ['Member'],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_DNS_RECORD_REMOVABLE_TYPES,
     'type': 'STRING', 'flags': 'V',
     'allowed_object_types': ['Member'],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_DHCP_SUPPORT,
     'type': 'ENUM', 'flags': '',
     'allowed_object_types': ['Member'],
     'list_values': [{'value': 'True'},
                     {'value': 'False'}],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_DNS_SUPPORT,
     'type': 'ENUM', 'flags': '',
     'allowed_object_types': ['Member'],
     'list_values': [{'value': 'True'},
                     {'value': 'False'}],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_RELAY_SUPPORT,
     'type': 'ENUM', 'flags': '',
     'allowed_object_types': ['Member'],
     'list_values': [{'value': 'True'},
                     {'value': 'False'}],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_USE_GM_FOR_DHCP,
     'type': 'ENUM', 'flags': '',
     'allowed_object_types': ['Member'],
     'list_values': [{'value': 'True'},
                     {'value': 'False'}],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_TENANT_NAME_PERSISTENCE,
     'type': 'ENUM', 'flags': '',
     'allowed_object_types': ['Member'],
     'list_values': [{'value': 'True'},
                     {'value': 'False'}],
     'comment': 'Grid Configuration'},

    {'name': EA_GRID_CONFIG_REPORT_GRID_SYNC_TIME,
     'type': 'ENUM', 'flags': '',
     'allowed_object_types': ['Member'],
     'list_values': [{'value': 'True'},
                     {'value': 'False'}],
     'comment': 'Grid Configuration'},

    # Grid Report
    {'name': EA_LAST_GRID_SYNC_TIME,
     'type': 'STRING', 'flags': 'CV',
     'allowed_object_types': ['Member'],
     'comment': 'Grid Sync Report'},

    # Mapping EAs
    {'name': EA_MAPPING_ADDRESS_SCOPE_ID,
     'type': 'STRING', 'flags': 'CGV',
     'allowed_object_types': ['NetworkView'],
     'comment': 'Mapping'},

    {'name': EA_MAPPING_ADDRESS_SCOPE_NAME,
     'type': 'STRING', 'flags': 'CGV',
     'allowed_object_types': ['NetworkView'],
     'comment': 'Mapping'},

    {'name': EA_MAPPING_TENANT_ID,
     'type': 'STRING', 'flags': 'CGV',
     'allowed_object_types': ['NetworkView'],
     'comment': 'Mapping'},

    {'name': EA_MAPPING_TENANT_NAME,
     'type': 'STRING', 'flags': 'CGV',
     'allowed_object_types': ['NetworkView'],
     'comment': 'Mapping'},

    {'name': EA_MAPPING_NETWORK_ID,
     'type': 'STRING', 'flags': 'CGV',
     'allowed_object_types': ['NetworkView'],
     'comment': 'Mapping'},

    {'name': EA_MAPPING_NETWORK_NAME,
     'type': 'STRING', 'flags': 'CGV',
     'allowed_object_types': ['NetworkView'],
     'comment': 'Mapping'},

    {'name': EA_MAPPING_SUBNET_ID,
     'type': 'STRING', 'flags': 'CGV',
     'allowed_object_types': ['NetworkView'],
     'comment': 'Mapping'},

    {'name': EA_MAPPING_SUBNET_CIDR,
     'type': 'STRING', 'flags': 'CGV',
     'allowed_object_types': ['NetworkView'],
     'comment': 'Mapping'},

    # Cloud Participation EA
    {'name': EA_CLOUD_ADAPTER_ID,
     'type': 'STRING', 'flags': 'CV',
     'comment': 'Cloud Participation'},

    # Member Registration EA
    {'name': EA_IS_CLOUD_MEMBER,
     'type': 'ENUM', 'flags': '',
     'allowed_object_types': ['Member'],
     'list_values': [{'value': 'True'},
                     {'value': 'False'}],
     'comment': 'Grid Membership'}
]

GRID_CONFIG_DEFAULTS = {
    EA_GRID_CONFIG_GRID_SYNC_SUPPORT: True,
    EA_GRID_CONFIG_GRID_SYNC_MINIMUM_WAIT_TIME: 60,
    EA_GRID_CONFIG_GRID_SYNC_MAXIMUM_WAIT_TIME: 300,
    EA_GRID_CONFIG_DEFAULT_NETWORK_VIEW_SCOPE: NETWORK_VIEW_SCOPE_SINGLE,
    EA_GRID_CONFIG_DEFAULT_NETWORK_VIEW: DEFAULT_NETWORK_VIEW,
    EA_GRID_CONFIG_DEFAULT_HOST_NAME_PATTERN: 'host-{ip_address}',
    EA_GRID_CONFIG_DEFAULT_DOMAIN_NAME_PATTERN: '{subnet_id}.cloud.global.com',
    EA_GRID_CONFIG_NS_GROUP: None,
    EA_GRID_CONFIG_DNS_VIEW: DEFAULT_DNS_VIEW,
    EA_GRID_CONFIG_NETWORK_TEMPLATE: None,
    EA_GRID_CONFIG_ADMIN_NETWORK_DELETION: False,
    EA_GRID_CONFIG_IP_ALLOCATION_STRATEGY: (
        IP_ALLOCATION_STRATEGY_FIXED_ADDRESS),
    EA_GRID_CONFIG_DNS_RECORD_BINDING_TYPES: [DNS_RECORD_TYPE_A,
                                              DNS_RECORD_TYPE_AAAA,
                                              DNS_RECORD_TYPE_PTR],
    EA_GRID_CONFIG_DNS_RECORD_UNBINDING_TYPES: [DNS_RECORD_TYPE_A,
                                                DNS_RECORD_TYPE_AAAA,
                                                DNS_RECORD_TYPE_PTR],
    EA_GRID_CONFIG_DNS_RECORD_REMOVABLE_TYPES: [DNS_RECORD_TYPE_A,
                                                DNS_RECORD_TYPE_AAAA,
                                                DNS_RECORD_TYPE_PTR,
                                                DNS_RECORD_TYPE_TXT],
    EA_GRID_CONFIG_DHCP_SUPPORT: False,
    EA_GRID_CONFIG_DNS_SUPPORT: False,
    EA_GRID_CONFIG_RELAY_SUPPORT: False,
    EA_GRID_CONFIG_USE_GM_FOR_DHCP: True,
    EA_GRID_CONFIG_TENANT_NAME_PERSISTENCE: False,
    EA_GRID_CONFIG_REPORT_GRID_SYNC_TIME: False,
    EA_LAST_GRID_SYNC_TIME: None
}



# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_serialization import jsonutils

from neutron import context
from neutron.tests.unit import testlib_api

from networking_infoblox.neutron.common import grid
from networking_infoblox.neutron.common import member
from networking_infoblox.neutron.common import utils
from networking_infoblox.neutron.db import infoblox_db as dbi

from networking_infoblox.tests import base


class GridMemberTestCase(base.TestCase, testlib_api.SqlTestCase):

    def setUp(self):

        super(GridMemberTestCase, self).setUp()
        self.ctx = context.get_admin_context()

        self.test_grid_config = grid.GridConfiguration(self.ctx)
        self.test_grid_config.gm_connector = mock.Mock()
        self.test_grid_config.grid_id = 100
        self.test_grid_config.grid_name = "Test Grid 1"
        self.test_grid_config.grid_master_host = '192.168.1.7'
        self.test_grid_config.grid_master_name = 'nios-7.2.0-master.com'
        self.test_grid_config.admin_user_name = 'admin'
        self.test_grid_config.admin_password = 'infoblox'
        self.test_grid_config.wapi_version = '2.2'

    def test_sync_grid(self):
        member_mgr = member.GridMemberManager(self.test_grid_config)
        member_mgr.sync_grid()

        grids = dbi.get_grids(self.ctx.session)
        self.assertEqual(1, len(grids))
        self.assertEqual(self.test_grid_config.grid_id, grids[0]['grid_id'])
        self.assertEqual(self.test_grid_config.grid_name,
                         grids[0]['grid_name'])
        expected_grid_connection = {
            "wapi_version": self.test_grid_config.wapi_version,
            "ssl_verify": self.test_grid_config.ssl_verify,
            "http_pool_connections":
                self.test_grid_config.http_pool_connections,
            "http_pool_maxsize": self.test_grid_config.http_pool_maxsize,
            "http_request_timeout":
                self.test_grid_config.http_request_timeout,
            "admin_user": {"name": self.test_grid_config.admin_user_name,
                           "password": self.test_grid_config.admin_password}
        }
        actual_grid_connection = jsonutils.loads(grids[0]['grid_connection'])
        self.assertEqual(expected_grid_connection, actual_grid_connection)
        self.assertEqual('ON', grids[0]['grid_status'])

        # change active grid to "Test Grid 2"
        new_active_grid_config = grid.GridConfiguration(self.ctx)
        new_active_grid_config.gm_connector = mock.Mock()
        new_active_grid_config.grid_id = 200
        new_active_grid_config.grid_name = "Test Grid 2"
        new_active_grid_config.grid_master_host = '192.168.1.8'
        new_active_grid_config.admin_user_name = 'admin'
        new_active_grid_config.admin_password = 'infoblox'
        new_active_grid_config.wapi_version = '1.4.2'
        member_mgr = member.GridMemberManager(new_active_grid_config)
        member_mgr.sync_grid()

        grids = dbi.get_grids(self.ctx.session)
        self.assertEqual(2, len(grids))
        self.assertEqual(self.test_grid_config.grid_id, grids[0]['grid_id'])
        self.assertEqual('OFF', grids[0]['grid_status'])

        self.assertEqual(new_active_grid_config.grid_id, grids[1]['grid_id'])
        self.assertEqual(new_active_grid_config.grid_name,
                         grids[1]['grid_name'])
        expected_grid_connection = {
            "wapi_version": new_active_grid_config.wapi_version,
            "ssl_verify": new_active_grid_config.ssl_verify,
            "http_pool_connections":
                new_active_grid_config.http_pool_connections,
            "http_pool_maxsize": new_active_grid_config.http_pool_maxsize,
            "http_request_timeout":
                new_active_grid_config.http_request_timeout,
            "admin_user": {"name": new_active_grid_config.admin_user_name,
                           "password": new_active_grid_config.admin_password}
        }
        actual_grid_connection = jsonutils.loads(grids[1]['grid_connection'])
        self.assertEqual(expected_grid_connection, actual_grid_connection)
        self.assertEqual('ON', grids[1]['grid_status'])

    def test_sync_member_without_cloud_support(self):
        # wapi version less than 2.0 indicates no cloud support
        self.test_grid_config.wapi_version = '1.4.2'
        member_mgr = member.GridMemberManager(self.test_grid_config)
        member_mgr.sync_grid()

        member_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_MEMBERS_WITHOUT_CLOUD)
        member_mgr._discover_members = mock.Mock()
        member_mgr._discover_members.return_value = member_json

        member_mgr._discover_member_licenses = mock.Mock()
        member_mgr._discover_member_licenses.return_value = None
        member_mgr.sync_members()

        members = dbi.get_members(self.ctx.session)
        self.assertEqual(len(member_json), len(members))
        for m in members:
            if self.test_grid_config.grid_master_host == m['member_ip']:
                self.assertEqual('GM', m['member_type'])
            else:
                self.assertEqual('REGULAR', m['member_type'])

    def test_sync_member_with_cloud_support_without_member_licenses(self):
        member_mgr = member.GridMemberManager(self.test_grid_config)
        member_mgr.sync_grid()

        member_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_MEMBERS_WITH_CLOUD)
        member_mgr._discover_members = mock.Mock()
        member_mgr._discover_members.return_value = member_json

        member_mgr._discover_member_licenses = mock.Mock()
        member_mgr._discover_member_licenses.return_value = None

        member_mgr.sync_members()

        members = dbi.get_members(self.ctx.session)
        self.assertEqual(len(member_json), len(members))
        for m in members:
            if self.test_grid_config.grid_master_host == m['member_ip']:
                self.assertEqual('GM', m['member_type'])
            else:
                self.assertEqual('REGULAR', m['member_type'])

    def test_sync_member_with_cloud_support_with_member_licenses(self):
        member_mgr = member.GridMemberManager(self.test_grid_config)
        member_mgr.sync_grid()

        member_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_MEMBERS_WITH_CLOUD)
        member_mgr._discover_members = mock.Mock()
        member_mgr._discover_members.return_value = member_json

        license_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_MEMBER_LICENSES)
        member_mgr._discover_member_licenses = mock.Mock()
        member_mgr._discover_member_licenses.return_value = license_json

        member_mgr.sync_members()

        # member_type is computed base on member licenses. To prove correctness
        # of member types, we need to first figure out which member has
        # 'CLOUD_API' license. A member with 'CLOUD_API' is CPM; otherwise
        # 'REGULAR'. Grid master member is 'GM'
        cloud_member_list = []
        cloud_lecensed_members = utils.find_in_list('type', ['CLOUD_API'],
                                                    license_json)
        member_hwids = utils.get_values_from_records('hwid',
                                                     cloud_lecensed_members)
        for m in member_json:
            member_id_arg = str(self.test_grid_config.grid_id) + m['host_name']
            member_id = utils.get_hash(member_id_arg)
            member_hwid = m['node_info'][0].get('hwid')
            if member_hwid in member_hwids:
                cloud_member_list.append(member_id)

        # now we know member licenses so get members in db
        members = dbi.get_members(self.ctx.session)
        self.assertEqual(len(member_json), len(members))
        # verify member types
        for m in members:
            if self.test_grid_config.grid_master_host == m['member_ip']:
                self.assertEqual('GM', m['member_type'])
            else:
                if m['member_id'] in cloud_member_list:
                    self.assertEqual('CPM', m['member_type'])
                else:
                    self.assertEqual('REGULAR', m['member_type'])



# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import netaddr
import sys

from oslo_log import log as logging
from oslo_utils import excutils

from neutron.i18n import _LE
from neutron.i18n import _LW
from neutron.ipam import driver
from neutron.ipam import exceptions as ipam_exc
from neutron.ipam import requests as ipam_req
from neutron.ipam import subnet_alloc
from neutron import manager

from infoblox_client import exceptions as ib_exc
from infoblox_client import objects as ib_objects

from networking_infoblox.ipam import requests
from networking_infoblox.neutron.common import constants as const
from networking_infoblox.neutron.common import context as ib_context
from networking_infoblox.neutron.common import dns
from networking_infoblox.neutron.common import exceptions as exc
from networking_infoblox.neutron.common import grid
from networking_infoblox.neutron.common import ipam
from networking_infoblox.neutron.db import infoblox_db as dbi


LOG = logging.getLogger(__name__)


def catch_ib_client_exception(f):
    @functools.wraps(f)
    def func(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except ib_exc.InfobloxException as e:
            raise exc.InfobloxClientException(msg=e), None, sys.exc_info()[2]
        except ValueError as e:
            raise exc.InfobloxValueError(msg=e), None, sys.exc_info()[2]
    return func


def rollback_wrapper(f):
    @functools.wraps(f)
    def rollback(*args, **kwargs):
        rollback_list = []
        try:
            return f(args[0], rollback_list, *args[1:], **kwargs)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("An exception occurred during %(action)s: "
                              "%(error)s"),
                          {'action': f.__name__, 'error': e})
                for ib_object in reversed(rollback_list):
                    try:
                        ib_object.delete()
                    except ib_exc.InfobloxException as e:
                        LOG.warning(_LW("Unable to delete %(obj)s due "
                                        "to error: %(error)s."),
                                    {'obj': ib_object, 'error': e})
    return rollback


class InfobloxPool(subnet_alloc.SubnetAllocator):
    """Infoblox Pool.

    InfobloxPool is responsible for subnet management in Infoblox backend.
    """

    @catch_ib_client_exception
    def __init__(self, subnetpool, context):
        super(InfobloxPool, self).__init__(subnetpool, context)
        self._plugin = manager.NeutronManager.get_plugin()
        self._grid_manager = grid.GridManager(self._context)
        self._grid_manager.get_config()
        self._grid_config = self._grid_manager.grid_config

    @catch_ib_client_exception
    def get_subnet(self, subnet_id):
        """Retrieve an IPAM subnet.

        :param subnet_id: Neutron subnet identifier
        :returns: a InfobloxSubnet instance
        """
        neutron_subnet = self._fetch_subnet(subnet_id)
        subnet_request = self._build_request_from_subnet(neutron_subnet)

        ib_cxt = ib_context.InfobloxContext(
            self._context,
            self._context.user_id,
            None,
            neutron_subnet,
            self._grid_config,
            plugin=self._plugin)

        ipam_controller = ipam.IpamSyncController(ib_cxt)
        ib_network = ipam_controller.get_subnet()
        if ib_network:
            return InfobloxSubnet(subnet_request, neutron_subnet, ib_network,
                                  ib_cxt)

    @staticmethod
    def _build_request_from_subnet(neutron_subnet):
        alloc_pools = None
        if neutron_subnet.get('allocation_pools'):
            alloc_pools = [netaddr.IPRange(pool['start'], pool['end'])
                           for pool in neutron_subnet['allocation_pools']]
        return ipam_req.SpecificSubnetRequest(
            neutron_subnet['tenant_id'],
            neutron_subnet['id'],
            neutron_subnet['cidr'],
            neutron_subnet['gateway_ip'],
            alloc_pools)

    def _fetch_subnet(self, subnet_id):
        return self._plugin.get_subnet(self._context, subnet_id)

    @catch_ib_client_exception
    @rollback_wrapper
    def allocate_subnet(self, rollback_list, subnet_request):
        """Create an IPAM subnet from the subnet request which contains cidr.

        Allocates a subnet to the Infoblox backend.
        :param subnet_request: instance of SubnetRequest child
        :returns: a InfobloxSubnet instance
        """
        # if subnetpool is defined, the request is AnySubnetRequest, so
        # we need to convert it to SpecificSubnetRequest calling
        # SubnetAllocator; however, calling this will not pass custom
        # parameters we defined so we need to get them back from the original
        # subnet_request.
        if self._subnetpool:
            orig_request = {
                'name': subnet_request.name,
                'network_id': subnet_request.network_id,
                'subnetpool_id': subnet_request.subnetpool_id,
                'enable_dhcp': subnet_request.enable_dhcp,
                'dns_nameservers': subnet_request.dns_nameservers}
            subnet = super(InfobloxPool, self).allocate_subnet(subnet_request)
            subnet_request = subnet.get_details()
            subnet_request.name = orig_request['name']
            subnet_request.network_id = orig_request['network_id']
            subnet_request.subnetpool_id = orig_request['subnetpool_id']
            subnet_request.enable_dhcp = orig_request['enable_dhcp']
            subnet_request.dns_nameservers = orig_request['dns_nameservers']

        # SubnetRequest must be SpecificSubnet at this point
        if not isinstance(subnet_request, ipam_req.SpecificSubnetRequest):
            raise ipam_exc.InvalidSubnetRequestType(
                subnet_type=type(subnet_request))

        neutron_subnet = self._build_subnet_from_request(subnet_request)
        ib_cxt = ib_context.InfobloxContext(
            self._context,
            self._context.user_id,
            None,
            neutron_subnet,
            self._grid_config,
            plugin=self._plugin)

        ipam_controller = ipam.IpamSyncController(ib_cxt)
        dns_controller = dns.DnsController(ib_cxt)

        ib_network = self._create_ib_network(rollback_list, ipam_controller)
        if ib_network:
            dns_controller.create_dns_zones(rollback_list)
            LOG.info("Created DNS zones.")

        return InfobloxSubnet(subnet_request, neutron_subnet, ib_network,
                              ib_cxt)

    @staticmethod
    def _build_subnet_from_request(subnet_request):
        return {'id': subnet_request.subnet_id,
                'name': subnet_request.name,
                'tenant_id': subnet_request.tenant_id,
                'network_id': subnet_request.network_id,
                'cidr': str(subnet_request.subnet_cidr),
                'ip_version': subnet_request.subnet_cidr.version,
                'subnetpool_id': subnet_request.subnetpool_id,
                'allocation_pools': subnet_request.allocation_pools,
                'gateway_ip': subnet_request.gateway_ip,
                'enable_dhcp': subnet_request.enable_dhcp,
                'dns_nameservers': subnet_request.dns_nameservers}

    @staticmethod
    def _create_ib_network(rollback_list, ipam_controller):
        ib_network = None
        retry = 1
        while True:
            try:
                LOG.info("Attempting to create ib network...")
                ib_network = ipam_controller.create_subnet(rollback_list)
                LOG.info("Successfully created ib network.")
                break
            except ib_exc.InfobloxMemberAlreadyAssigned:
                LOG.info("ib network creation failed due to member assignment "
                         "issue, tried %s time(s)", retry)
                retry += 1
                if retry > const.MEMBER_RESERVATION_RETRY:
                    raise exc.InfobloxCannotCreateSubnet(
                        reason="The grid has no more member available "
                               "to server the current subnet.")
        return ib_network

    @catch_ib_client_exception
    @rollback_wrapper
    def update_subnet(self, rollback_list, subnet_request):
        """Update IPAM Subnet.

        Updates allocation pools, dns zones, or EAs for the subnet in the
        Infoblox backend.
        """
        neutron_subnet = self._build_subnet_from_request(subnet_request)
        ib_network = self._get_ib_network(neutron_subnet['id'],
                                          neutron_subnet['ip_version'])
        if not ib_network:
            raise exc.InfobloxCannotFindSubnet(subnet_id=neutron_subnet['id'],
                                               cidr=neutron_subnet['cidr'])

        ib_cxt = ib_context.InfobloxContext(
            self._context,
            self._context.user_id,
            None,
            neutron_subnet,
            self._grid_config,
            plugin=self._plugin,
            ib_network=ib_network)

        ipam_controller = ipam.IpamSyncController(ib_cxt)
        dns_controller = dns.DnsController(ib_cxt)

        ipam_controller.update_subnet_allocation_pools(rollback_list)

        if self._is_new_zone_required(neutron_subnet, ib_network):
            # subnet name is used in the domain suffix pattern and the name
            # has been changed; we need to create new zones.
            dns_controller.create_dns_zones(rollback_list)

        ipam_controller.update_subnet_details(ib_network)

    def _is_new_zone_required(self, subnet, ib_network):
        pattern = self._grid_config.default_domain_name_pattern
        subnet_used = '{subnet_name}' in pattern
        if subnet_used:
            new_subnet_name = subnet.get('name')
            old_subnet_name = ib_network.extattrs.get(const.EA_SUBNET_NAME)
            return (old_subnet_name is not None and
                    new_subnet_name is not None and
                    old_subnet_name != new_subnet_name)
        return False

    @catch_ib_client_exception
    def remove_subnet(self, subnet_id):
        """Remove IPAM Subnet.

        Removes a subnet from the Infoblox backend.
        """
        ib_network = self._get_ib_network(subnet_id)
        if not ib_network:
            return

        neutron_subnet = self._build_subnet_from_ib_network(ib_network)
        ib_cxt = ib_context.InfobloxContext(
            self._context,
            self._context.user_id,
            None,
            neutron_subnet,
            self._grid_config,
            plugin=self._plugin,
            ib_network=ib_network)

        ipam_controller = ipam.IpamSyncController(ib_cxt)
        dns_controller = dns.DnsController(ib_cxt)

        ipam_controller.delete_subnet(ib_network)
        dns_controller.delete_dns_zones(ib_network=ib_network)

    def _get_ib_network(self, subnet_id, ip_version=None):
        db_netviews = dbi.get_network_view_by_mapping(
            self._context.session,
            grid_id=self._grid_config.grid_id,
            subnet_id=subnet_id)
        if not db_netviews:
            return None

        network_view = db_netviews[0].network_view
        ea = ib_objects.EA({'Subnet ID': subnet_id})
        ib_network = None
        # TODO(hhwang): this should be replaced to use get_network_by_subnet_id
        # api when the following issue is resolved:
        # https://github.com/infobloxopen/infoblox-client/issues/58
        if ip_version == 4 or ip_version is None:
            ib_network = ib_objects.NetworkV4.search(
                self._grid_config.gm_connector,
                network_view=network_view,
                search_extattrs=ea)
        if ip_version == 6 or ib_network is None:
            ib_network = ib_objects.NetworkV6.search(
                self._grid_config.gm_connector,
                network_view=network_view,
                search_extattrs=ea)
        return ib_network

    @staticmethod
    def _build_subnet_from_ib_network(ib_network):
        subnet = dict()
        subnet['id'] = ib_network.extattrs.get(const.EA_SUBNET_ID)
        subnet['name'] = ib_network.extattrs.get(const.EA_SUBNET_NAME)
        subnet['network_id'] = ib_network.extattrs.get(const.EA_NETWORK_ID)
        subnet['tenant_id'] = ib_network.extattrs.get(const.EA_TENANT_ID)
        subnet['cidr'] = ib_network.network
        subnet['ip_version'] = ib_network.ip_version
        return subnet

    def get_subnet_request_factory(self):
        """Returns InfobloxSubnetRequestFactory"""
        return requests.InfobloxSubnetRequestFactory

    def get_address_request_factory(self):
        """Returns InfobloxAddressRequestFactory"""
        return requests.InfobloxAddressRequestFactory


class InfobloxSubnet(driver.Subnet):
    """Infoblox IPAM subnet."""

    def __init__(self, subnet_details, neutron_subnet, ib_network, ib_cxt):
        self._validate_subnet_data(subnet_details)
        self._subnet_details = subnet_details
        self._neutron_subnet = neutron_subnet
        self._ib_network = ib_network
        self._ib_cxt = ib_cxt

    @staticmethod
    def _validate_subnet_data(subnet_details):
        if not isinstance(subnet_details, ipam_req.SpecificSubnetRequest):
            raise ValueError("Subnet details should be passed as "
                             "SpecificSubnetRequest")

    @catch_ib_client_exception
    def allocate(self, address_request):
        """Allocate an IP address based on the request passed in.

        :param address_request: Specifies what to allocate.
        :type address_request: A subclass of AddressRequest
        :returns: A netaddr.IPAddress
        """
        ipam_controller = ipam.IpamSyncController(self._ib_cxt)
        dns_controller = dns.DnsController(self._ib_cxt)

        if isinstance(address_request, ipam_req.SpecificAddressRequest):
            allocated_ip = ipam_controller.allocate_specific_ip(
                str(address_request.address),
                address_request.mac,
                address_request.port_id,
                address_request.tenant_id,
                address_request.device_id,
                address_request.device_owner)
        else:
            allocated_ip = ipam_controller.allocate_ip_from_pool(
                self._neutron_subnet.get('id'),
                self._neutron_subnet.get('allocation_pools'),
                address_request.mac,
                address_request.port_id,
                address_request.tenant_id,
                address_request.device_id,
                address_request.device_owner)

        if allocated_ip and address_request.device_owner:
            # we can deal with instance name as hostname in the ipam agent.
            instance_name = None
            try:
                dns_controller.bind_names(allocated_ip,
                                          instance_name,
                                          address_request.port_id,
                                          address_request.tenant_id,
                                          address_request.device_id,
                                          address_request.device_owner)
            except Exception:
                with excutils.save_and_reraise_exception():
                    ipam_controller.deallocate_ip(allocated_ip)

        return allocated_ip

    @catch_ib_client_exception
    def deallocate(self, address):
        """Deallocate previously allocated address.

        :param address: The address to deallocate.
        :type address: A subclass of netaddr.IPAddress or convertible to one.
        :returns: None
        """
        ip_addr = str(address)
        address_request = self._build_address_request_from_ib_address(ip_addr)
        if not address_request:
            return

        ipam_controller = ipam.IpamSyncController(self._ib_cxt)
        dns_controller = dns.DnsController(self._ib_cxt)

        ipam_controller.deallocate_ip(ip_addr)
        dns_controller.unbind_names(ip_addr,
                                    None,
                                    address_request.port_id,
                                    address_request.tenant_id,
                                    address_request.device_id,
                                    address_request.device_owner)

    def _build_address_request_from_ib_address(self, ip_address):
        connector = self._ib_cxt.connector
        netview = self._ib_cxt.mapping.network_view
        dns_view = self._ib_cxt.mapping.dns_view

        ib_address = ib_objects.FixedAddress.search(connector,
                                                    network_view=netview,
                                                    ip=ip_address)
        if not ib_address:
            ib_address = ib_objects.HostRecord.search(connector,
                                                      view=dns_view,
                                                      ip=ip_address)
            if not ib_address:
                return None

        addr_req = ipam_req.AddressRequest()
        addr_req.port_id = ib_address.extattrs.get(const.EA_PORT_ID)
        addr_req.tenant_id = ib_address.extattrs.get(const.EA_TENANT_ID)
        addr_req.device_id = ib_address.extattrs.get(const.EA_PORT_DEVICE_ID)
        addr_req.device_owner = ib_address.extattrs.get(
            const.EA_PORT_DEVICE_OWNER)
        return addr_req

    def get_details(self):
        """Return subnet detail as a SpecificSubnetRequest.

        :returns: An instance of SpecificSubnetRequest with the subnet detail.
        """
        return self._subnet_details



# Copyright 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import six

from oslo_log import log as logging

from networking_infoblox.neutron.common import constants as const


LOG = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class IPAllocator(object):
    """Base class for ip allocation actions"""

    DEFAULT_OPTIONS = {
        'use_host_record': True,
        'configure_for_dhcp': True,
        'configure_for_dns': True,
        'dns_record_binding_types': [],
        'dns_record_unbinding_types': [],
        'dns_record_removable_types': []}

    def __new__(cls, ib_obj_manager, options):
        cls._validate_and_set_default_options(options)
        return super(IPAllocator, cls).__new__(
            cls._get_ip_allocator_class(options))

    def __init__(self, ib_obj_manager, options):
        self.manager = ib_obj_manager
        self.opts = options

    @classmethod
    def _validate_and_set_default_options(cls, options):
        if not isinstance(options, dict):
            raise ValueError('Options should be passed as dict')
        for key in cls.DEFAULT_OPTIONS:
            if key not in options:
                options[key] = cls.DEFAULT_OPTIONS[key]

    @staticmethod
    def _get_ip_allocator_class(options):
        if options['use_host_record']:
            return HostRecordIPAllocator
        else:
            return FixedAddressIPAllocator

    @abc.abstractmethod
    def allocate_ip_from_range(self, network_view, dns_view, zone_auth,
                               hostname, mac, first_ip, last_ip,
                               extattrs=None):
        pass

    @abc.abstractmethod
    def allocate_given_ip(self, network_view, dns_view, zone_auth,
                          hostname, mac, ip, extattrs=None):
        pass

    @abc.abstractmethod
    def deallocate_ip(self, network_view, dns_view_name, ip):
        pass

    @abc.abstractmethod
    def bind_names(self, network_view, dns_view, ip, name, extattrs):
        pass

    @abc.abstractmethod
    def unbind_names(self, network_view, dns_view, ip, name, extattrs):
        pass


class HostRecordIPAllocator(IPAllocator):

    def bind_names(self, network_view, dns_view, ip, name, extattrs):
        # See OPENSTACK-181. In case hostname already exists on NIOS, update
        # host record which contains that hostname with the new IP address
        # rather than creating a separate host record object
        reserved_hostname_hr = self.manager.find_hostname(dns_view, name, ip)
        reserved_ip_hr = self.manager.get_host_record(dns_view, ip)

        if (reserved_hostname_hr and reserved_ip_hr and
                reserved_hostname_hr.ref == reserved_ip_hr.ref):
            reserved_hostname_hr.extattrs = extattrs
            reserved_hostname_hr.update()
            return

        if reserved_hostname_hr:
            for hr_ip in reserved_ip_hr.ips:
                if hr_ip == ip:
                    reserved_ip_hr.delete()
                    reserved_hostname_hr.extattrs = extattrs
                    self.manager.add_ip_to_record(
                        reserved_hostname_hr, ip, hr_ip.mac)
                    break
        else:
            self.manager.bind_name_with_host_record(
                dns_view, ip, name, extattrs)

    def unbind_names(self, network_view, dns_view, ip, name, extattrs):
        # Nothing to delete, all will be deleted together with host record.
        pass

    def allocate_ip_from_range(self, network_view, dns_view,
                               zone_auth, hostname, mac, first_ip, last_ip,
                               extattrs=None):
        use_dhcp = self.opts['configure_for_dhcp']
        use_dns = self.opts['configure_for_dns']
        fqdn = hostname + '.' + zone_auth
        host_record = self.manager.find_hostname(
            dns_view, fqdn, first_ip)
        if host_record:
            hr = self.manager.add_ip_to_host_record_from_range(
                host_record, network_view, mac, first_ip, last_ip, use_dhcp)
        else:
            hr = self.manager.create_host_record_from_range(
                dns_view, network_view, zone_auth, hostname, mac,
                first_ip, last_ip, extattrs, use_dhcp, use_dns)
        return hr.ip[-1].ip

    def allocate_given_ip(self, network_view, dns_view, zone_auth,
                          hostname, mac, ip, extattrs=None):
        use_dhcp = self.opts['configure_for_dhcp']
        use_dns = self.opts['configure_for_dns']
        hr = self.manager.create_host_record_for_given_ip(
            dns_view, zone_auth, hostname, mac, ip, extattrs, use_dhcp,
            use_dns)
        return hr.ip[-1].ip

    def deallocate_ip(self, network_view, dns_view_name, ip):
        host_record = self.manager.get_host_record(dns_view_name, ip)
        if host_record:
            if len(host_record.ip) > 1:
                self.manager.delete_ip_from_host_record(host_record, ip)
            else:
                host_record.delete()


class FixedAddressIPAllocator(IPAllocator):

    def bind_names(self, network_view, dns_view, ip, name, extattrs):
        bind_cfg = self.opts['dns_record_binding_types']
        device_owner = extattrs.get(const.EA_PORT_DEVICE_OWNER)
        if device_owner in const.NEUTRON_FLOATING_IP_DEVICE_OWNERS:
            self.manager.update_fixed_address_eas(network_view, ip, extattrs)
            if self.opts['configure_for_dns']:
                self.manager.update_dns_record_eas(dns_view, ip, extattrs)
        if bind_cfg and self.opts['configure_for_dns']:
            self.manager.bind_name_with_record_a(
                dns_view, ip, name, bind_cfg, extattrs)

    def unbind_names(self, network_view, dns_view, ip, name, extattrs):
        unbind_cfg = self.opts['dns_record_unbinding_types']
        if unbind_cfg and self.opts['configure_for_dns']:
            self.manager.unbind_name_from_record_a(
                dns_view, ip, name, unbind_cfg)

    def allocate_ip_from_range(self, network_view, dns_view,
                               zone_auth, hostname, mac, first_ip, last_ip,
                               extattrs=None):
        fa = self.manager.create_fixed_address_from_range(
            network_view, mac, first_ip, last_ip, extattrs)
        return fa.ip

    def allocate_given_ip(self, network_view, dns_view, zone_auth,
                          hostname, mac, ip, extattrs=None):
        fa = self.manager.create_fixed_address_for_given_ip(
            network_view, mac, ip, extattrs)
        return fa.ip

    def deallocate_ip(self, network_view, dns_view_name, ip):
        delete_cfg = self.opts['dns_record_removable_types']
        if delete_cfg and self.opts['configure_for_dns']:
            self.manager.unbind_name_from_record_a(dns_view_name, ip,
                                                   None, delete_cfg)
        self.manager.delete_fixed_address(network_view, ip)



# Copyright (c) 2015 Infoblox Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add_dns_view_column_to_network_views

Revision ID: 256b90dd9824
Revises: 172ace2194db
Create Date: 2016-02-01 13:32:13.135126

"""

# revision identifiers, used by Alembic.
revision = '256b90dd9824'
down_revision = '172ace2194db'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        'infoblox_network_views',
        sa.Column('dns_view', sa.String(length=255), nullable=True)
    )
    op.add_column(
        'infoblox_network_views',
        sa.Column('internal_network_view', sa.String(length=255),
                  nullable=False)
    )
    op.add_column(
        'infoblox_network_views',
        sa.Column('internal_dns_view', sa.String(length=255), nullable=True)
    )
    op.create_index(
        op.f('ix_infoblox_network_views_grid_id_internal_network_view'),
        'infoblox_network_views',
        ['grid_id', 'internal_network_view'],
        unique=False)



# Copyright (c) 2015 Infoblox Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""member_sync_improvement

Revision ID: 4d0bb1d080f8
Revises: 422e067b7d36
Create Date: 2016-02-22 17:57:57.245145

"""

# revision identifiers, used by Alembic.
revision = '4d0bb1d080f8'
down_revision = '422e067b7d36'

from alembic import op
import sqlalchemy as sa
from sqlalchemy import update

from networking_infoblox.neutron.common import utils


# simple models for infoblox_grids and infoblox_grid_members with only the
# fields needed for the migration
infoblox_grids = sa.Table(
    'infoblox_grids',
    sa.MetaData(),
    sa.Column('grid_id', sa.Integer(), nullable=False),
    sa.Column('grid_status', sa.String(length=6), nullable=False),
    sa.Column('gm_id', sa.String(length=32), nullable=False,
              default=utils.get_hash()))

infoblox_grid_members = sa.Table(
    'infoblox_grid_members',
    sa.MetaData(),
    sa.Column('member_id', sa.String(length=32), nullable=False),
    sa.Column('grid_id', sa.Integer(), nullable=False),
    sa.Column('member_ip', sa.String(length=15), nullable=True),
    sa.Column('member_ipv6', sa.String(length=64), nullable=True),
    sa.Column('member_type', sa.String(length=6), nullable=False))


def upgrade():
    op.add_column(
        'infoblox_grids',
        sa.Column('gm_id', sa.String(length=32), nullable=False,
                  default=utils.get_hash()))

    op.drop_constraint(
        constraint_name='uniq_infoblox_grid_members_grid_id_member_name',
        table_name='infoblox_grid_members',
        type_='unique')

    op.add_column(
        'infoblox_grid_members',
        sa.Column('member_dhcp_ip', sa.String(length=15), nullable=True))
    op.add_column(
        'infoblox_grid_members',
        sa.Column('member_dhcp_ipv6', sa.String(length=64), nullable=True))
    op.add_column(
        'infoblox_grid_members',
        sa.Column('member_dns_ip', sa.String(length=15), nullable=True))
    op.add_column(
        'infoblox_grid_members',
        sa.Column('member_dns_ipv6', sa.String(length=64), nullable=True))
    op.add_column(
        'infoblox_grid_members',
        sa.Column('member_wapi', sa.String(length=255), nullable=True))

    update_gm_ids()


def update_gm_ids():
    session = sa.orm.Session(bind=op.get_bind())
    q = session.query(infoblox_grid_members).filter(
        infoblox_grid_members.c.member_type == 'GM')
    gm_rows = q.all()
    for gm_row in gm_rows:
        u = update(infoblox_grids, infoblox_grids.c.grid_id == gm_row.grid_id)
        session.execute(u, {'gm_id': gm_row.member_id})
    session.commit()



# Copyright (c) 2015 Infoblox Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add_infoblox_ipam_driver

Revision ID: 172ace2194db
Revises: None
Create Date: 2015-09-10 13:43:13.135126

"""

# revision identifiers, used by Alembic.
revision = '172ace2194db'
down_revision = 'start_networking_infoblox'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'infoblox_grids',
        sa.Column('grid_id', sa.Integer(), nullable=False),
        sa.Column('grid_name', sa.String(length=128), nullable=True),
        sa.Column('grid_connection', sa.String(length=1024), nullable=False),
        sa.Column('grid_status', sa.String(length=6), nullable=False),
        sa.PrimaryKeyConstraint('grid_id')
    )

    op.create_table(
        'infoblox_grid_members',
        sa.Column('member_id', sa.String(length=32), nullable=False),
        sa.Column('grid_id', sa.Integer(), nullable=False),
        sa.Column('member_name', sa.String(length=255), nullable=False),
        sa.Column('member_ip', sa.String(length=15), nullable=True),
        sa.Column('member_ipv6', sa.String(length=64), nullable=True),
        sa.Column('member_type', sa.String(length=12), nullable=False),
        sa.Column('member_status', sa.String(length=16), nullable=False),
        sa.ForeignKeyConstraint(['grid_id'],
                                ['infoblox_grids.grid_id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('member_id'),
        sa.UniqueConstraint(
            'grid_id', 'member_name',
            name='uniq_infoblox_grid_members_grid_id_member_name'),
        sa.UniqueConstraint(
            'member_id', 'grid_id', 'member_ip',
            name='uniq_infoblox_grid_members_member_id_grid_id_member_ip'),
        sa.UniqueConstraint(
            'member_id', 'grid_id', 'member_ipv6',
            name='uniq_infoblox_grid_members_member_id_grid_id_member_ipv6'),
        sa.Index(
            'ix_infoblox_grid_members_grid_id_member_status',
            'grid_id', 'member_status')
    )

    op.create_table(
        'infoblox_operations',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('op_type', sa.String(length=48), nullable=False),
        sa.Column('op_value', sa.String(length=255), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint(
            'op_type',
            name='uniq_infoblox_operations_op_type')
    )

    op.create_table(
        'infoblox_network_views',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('network_view', sa.String(length=255), nullable=False),
        sa.Column('grid_id', sa.Integer(), nullable=False),
        sa.Column('authority_member_id', sa.String(length=32), nullable=False),
        sa.Column('shared', sa.Boolean(), default=False, nullable=False),
        sa.ForeignKeyConstraint(['authority_member_id'],
                                ['infoblox_grid_members.member_id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint(
            'network_view', 'grid_id',
            name='uniq_infoblox_network_views_network_view_grid_id'),
        sa.UniqueConstraint(
            'network_view', 'authority_member_id',
            name='uniq_infoblox_network_views_network_view_authority_member_id'
        )
    )

    op.create_table(
        'infoblox_network_view_mapping',
        sa.Column('network_view_id', sa.String(length=36), nullable=False),
        sa.Column('network_id', sa.String(length=36), nullable=False),
        sa.Column('subnet_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['network_id'],
                                ['networks.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['network_view_id'],
                                ['infoblox_network_views.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('network_view_id', 'network_id', 'subnet_id')
    )

    op.create_table(
        'infoblox_mapping_conditions',
        sa.Column('network_view_id', sa.String(length=36), nullable=False),
        sa.Column('neutron_object_name', sa.String(length=48), nullable=False),
        sa.Column('neutron_object_value', sa.String(length=255),
                  nullable=False),
        sa.ForeignKeyConstraint(['network_view_id'],
                                ['infoblox_network_views.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('network_view_id', 'neutron_object_name',
                                'neutron_object_value')
    )

    op.create_table(
        'infoblox_mapping_members',
        sa.Column('network_view_id', sa.String(length=36), nullable=False),
        sa.Column('member_id', sa.String(length=32), nullable=False),
        sa.Column('mapping_relation', sa.String(length=24), nullable=False),
        sa.ForeignKeyConstraint(['network_view_id'],
                                ['infoblox_network_views.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['member_id'],
                                ['infoblox_grid_members.member_id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('network_view_id', 'member_id')
    )

    op.create_table(
        'infoblox_service_members',
        sa.Column('network_view_id', sa.String(length=36), nullable=False),
        sa.Column('member_id', sa.String(length=48), nullable=False),
        sa.Column('service', sa.String(length=12), nullable=False),
        sa.ForeignKeyConstraint(['network_view_id'],
                                ['infoblox_network_views.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['member_id'],
                                ['infoblox_grid_members.member_id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('network_view_id', 'member_id', 'service'),
        sa.UniqueConstraint(
            'member_id', 'service',
            name='uniq_infoblox_service_members_member_id_service')
    )

    op.create_table(
        'infoblox_tenants',
        sa.Column('tenant_id', sa.String(64),
                  nullable=False, primary_key=True),
        sa.Column('tenant_name', sa.String(64), nullable=False),
    )



#!/usr/bin/env python
# Copyright 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import sys

from oslo_config import cfg
from oslo_log import log as logging

from neutron.common import config as common_config
from neutron import context as neutron_context
from neutron.i18n import _LE
from neutron.i18n import _LW

from neutronclient.v2_0 import client as neutron_client
from novaclient import client as nova_client

from infoblox_client import exceptions as ib_exc

from networking_infoblox.neutron.common import config
from networking_infoblox.neutron.common import context as ib_context
from networking_infoblox.neutron.common import dns
from networking_infoblox.neutron.common import grid
from networking_infoblox.neutron.common import ipam
from networking_infoblox.neutron.common import utils
from networking_infoblox.neutron.db import infoblox_db as dbi


LOG = logging.getLogger(__name__)

NOVA_API_VERSION = '2'
DEFAULT_CONFIG_FILES = ['/etc/neutron/neutron.conf']


def main():
    cfg.CONF(args=sys.argv[1:],
             default_config_files=DEFAULT_CONFIG_FILES)
    common_config.setup_logging()
    config.register_infoblox_ipam_opts(cfg.CONF)
    grid_id = cfg.CONF.infoblox.cloud_data_center_id
    config.register_infoblox_grid_opts(cfg.CONF, grid_id)

    credentials = get_credentials()
    if not credentials:
        print("\nYou must provide an admin user credentials in the shell "
              "environment.\nPlease export variables such as env[OS_USERNAME],"
              " env[OS_PASSWORD], env[OS_AUTH_URL], env[OS_TENANT_NAME], and "
              "env[OS_REGION_NAME]\n")
        return

    context = neutron_context.get_admin_context()

    grid_manager = grid.GridManager(context)
    grid_manager.sync(force_sync=True)

    sync_neutron_to_infoblox(context, credentials, grid_manager)


def get_credentials():
    d = dict()
    if ('OS_USERNAME' in os.environ and
            'OS_PASSWORD' in os.environ and
            'OS_AUTH_URL' in os.environ and
            'OS_TENANT_NAME' in os.environ and
            'OS_REGION_NAME' in os.environ):
        d['username'] = os.environ['OS_USERNAME']
        d['password'] = os.environ['OS_PASSWORD']
        d['auth_url'] = os.environ['OS_AUTH_URL']
        d['tenant_name'] = os.environ['OS_TENANT_NAME']
        d['region_name'] = os.environ['OS_REGION_NAME']
    return d


def sync_neutron_to_infoblox(context, credentials, grid_manager):
    """Sync neutron objects to Infoblox grid

    Prerequisites:
        1. network views to sync must have "Cloud Adapter ID" EA set.
        2. infoblox agent sync should have been processed and updated members
           and network views.
    """
    LOG.info("Starting migration...\n")

    grid_config = grid_manager.grid_config
    grid_id = grid_config.grid_id
    session = context.session

    neutron_api = neutron_client.Client(**credentials)
    payload = neutron_api.list_networks()
    networks = payload['networks']
    if not networks:
        LOG.info("No network exists...Exiting...")
        return

    payload = neutron_api.list_subnets()
    subnets = payload['subnets']
    if not subnets:
        LOG.info("No subnet exists...Exiting...")
        return

    payload = neutron_api.list_ports()
    ports = payload['ports']

    nova_api = nova_client.Client(NOVA_API_VERSION,
                                  credentials['username'],
                                  credentials['password'],
                                  credentials['tenant_name'],
                                  credentials['auth_url'])
    instance_names_by_instance_id = dict()
    instance_names_by_floating_ip = dict()
    for server in nova_api.servers.list(search_opts={'all_tenants': 1}):
        instance_names_by_instance_id[server.id] = server.name
        floating_ips = []
        for net in server.addresses:
            floating_ips += [ip['addr'] for ip in server.addresses[net]
                             if ip['OS-EXT-IPS:type'] == 'floating']
        for fip in floating_ips:
            instance_names_by_floating_ip[fip] = server.name

    user_id = context.user_id or neutron_api.httpclient.auth_ref['user']['id']
    user_tenant_id = context.tenant_id or neutron_api.httpclient.auth_tenant_id

    should_exit = False

    # sync subnets
    for subnet in subnets:
        subnet_id = subnet['id']
        subnet_name = subnet['name']
        network_id = subnet['network_id']
        network = utils.find_one_in_list('id', network_id, networks)
        if not network:
            LOG.warning("network (%s) is not found. Skipping subnet (%s)",
                        network_id, subnet_id)
            continue

        network_name = network['name']
        db_mapped_netview = dbi.get_network_view_by_mapping(
            session,
            grid_id=grid_id,
            network_id=network_id,
            subnet_id=subnet_id)
        if db_mapped_netview:
            LOG.info("Mapping found for network (%s), subnet (%s)",
                     network_name, subnet_name)
            continue

        ib_cxt = ib_context.InfobloxContext(context, user_id,
                                            network, subnet,
                                            grid_config,
                                            plugin=neutron_api)
        ipam_controller = ipam.IpamSyncController(ib_cxt)
        dns_controller = dns.DnsController(ib_cxt)

        rollback_list = []
        try:
            ib_network = ipam_controller.create_subnet(rollback_list)
            if ib_network:
                dns_controller.create_dns_zones(rollback_list)
            LOG.info("Created network (%s), subnet (%s)",
                     network_name, subnet_name)
        except Exception as e:
            LOG.error(_LE("Error occurred: %(error)s"), {'error': e})
            for ib_obj in reversed(rollback_list):
                try:
                    ib_obj.delete()
                except ib_exc.InfobloxException as e:
                    LOG.warning(_LW("Unable to delete %(obj)s due to "
                                    "error: %(error)s."),
                                {'obj': ib_obj, 'error': e})
            should_exit = True
            break

    if should_exit:
        LOG.info("Exiting due to the error in creating subnet...")
        return

    # sync ports
    for port in ports:
        port_id = port['id']
        port_mac_address = port['mac_address']
        tenant_id = port.get('tenant_id') or user_tenant_id
        network_id = port['network_id']
        device_owner = port['device_owner']
        device_id = port['device_id']

        instance_name = (instance_names_by_instance_id[device_id]
                         if device_id in instance_names_by_instance_id
                         else None)

        network = utils.find_one_in_list('id', network_id, networks)
        if not network:
            LOG.error("network (%s) not found", network_id)
            break

        for ip_set in port.get('fixed_ips'):
            subnet_id = ip_set['subnet_id']
            ip_address = ip_set['ip_address']
            LOG.info("Adding port for %s: %s...", device_owner, ip_address)

            subnet = utils.find_one_in_list('id', subnet_id, subnets)
            if not subnet:
                should_exit = True
                LOG.error("subnet (%s) not found", subnet_id)
                break

            ib_cxt = ib_context.InfobloxContext(context, user_id,
                                                network, subnet,
                                                grid_config,
                                                plugin=neutron_api)
            connector = ib_cxt.connector
            netview = ib_cxt.mapping.network_view

            search_fields = {
                'network_view': netview,
                'ip_address': ip_address
            }
            obj_type = ('ipv4address'if utils.get_ip_version(ip_address) == 4
                        else 'ipv6address')
            ib_address = connector.get_object(obj_type,
                                              search_fields,
                                              return_fields=['objects'],
                                              force_proxy=True)
            if ib_address and ib_address[0]['objects']:
                LOG.info("%s is found...no need to create", ip_address)
                continue

            ipam_controller = ipam.IpamSyncController(ib_cxt)
            dns_controller = dns.DnsController(ib_cxt)

            # for a floating ip port, check for its association.
            # if associated, then port info needs to be the associated port,
            # not the floating ip port because the associated port contains
            # actual attached device info
            is_floating_ip = False
            if ip_address in instance_names_by_floating_ip:
                db_floatingip = dbi.get_floatingip_by_ip_address(session,
                                                                 ip_address)
                db_port = dbi.get_port_by_id(session,
                                             db_floatingip.fixed_port_id)
                port_id = db_port.id
                tenant_id = db_port.tenant_id
                device_id = db_port.device_id
                device_owner = db_port.device_owner
                instance_name = instance_names_by_floating_ip[ip_address]
                is_floating_ip = True

            allocated_ip = ipam_controller.allocate_specific_ip(
                ip_address,
                port_mac_address,
                port_id,
                tenant_id,
                device_id,
                device_owner)
            if allocated_ip and device_owner:
                try:
                    dns_controller.bind_names(
                        allocated_ip,
                        instance_name,
                        port_id,
                        tenant_id,
                        device_id,
                        device_owner,
                        is_floating_ip)
                except Exception as e:
                    should_exit = True
                    LOG.error("Unable to allocate ip (%s): %s", ip_address, e)
                    ipam_controller.deallocate_ip(allocated_ip)
                    break

            LOG.info("Allocated %s", ip_address)

        if should_exit:
            LOG.info("Existing due to error in port creation...")
            break

    LOG.info("Ending migration...")


if __name__ == "__main__":
    main()
